use core::ptr;
use cstr_core::{cstr, CStr};
use crate::error::Error;
use super::ffi;



/// Storage error type.
#[derive(Copy, Clone, Debug, PartialEq, Eq)]
pub enum StorageError {
    /// Data is too long or not appropriate for the operation.
    InvalidData,
    /// Write failed.
    WriteFailed,
    /// Read failed.
    ReadFailed,
    /// Failed to perform the get-and-increment operation on a counter.
    CounterFailed,
}

impl From<StorageError> for Error {
    fn from(err: StorageError) -> Self {
        match err {
            StorageError::InvalidData => Error::ValueError(cstr!("Invalid data for storage")),
            StorageError::WriteFailed => Error::ValueError(cstr!("Storage write failed")),
            StorageError::ReadFailed => Error::ValueError(cstr!("Storage read failed")),
            StorageError::CounterFailed => {
                Error::ValueError(cstr!("Retrieving counter value failed"))
            }
        }
    }
}
pub type StorageResult<T> = Result<T, StorageError>;

/// Result of PIN delay callback.
#[derive(Copy, Clone, Debug, PartialEq, Eq)]
pub enum PinCallbackResult {
    /// Continue waiting for PIN unlock.
    Continue,
    /// Abort the unlock attempt before time is up.
    Abort,
}

/// PIN delay callback function type.
/// The storage layer will call this function while the PIN timeout is in
/// progress. This is useful for showing UI progress bar.
/// `wait` is the total number of seconds waiting.
/// `progress` is a value between 0 and 1000, where 1000 indicates 100%.
/// `message` is a message to show to the user.
pub type PinDelayCallback = fn(wait: u32, progress: u32, message: &str) -> PinCallbackResult;

/// Static reference to the currently set PIN callback function.
static mut PIN_UI_CALLBACK: Option<PinDelayCallback> = None;

/// C-compatible wrapper for the Rust callback.
unsafe extern "C" fn callback_wrapper(
    wait: u32,
    progress: u32,
    message: *const cty::c_char,
) -> ffi::secbool {
    let message = unsafe { CStr::from_ptr(message as _) };
    let result = unsafe {
        PIN_UI_CALLBACK
            .map(|c| c(wait, progress, message.to_str().unwrap_or("")))
            .unwrap_or(PinCallbackResult::Continue)
    };
    if matches!(result, PinCallbackResult::Abort) {
        ffi::sectrue
    } else {
        ffi::secfalse
    }
}

pub fn storage_init(callback: PinDelayCallback) {
    unsafe {
        PIN_UI_CALLBACK = Some(callback);
        ffi::storage_init(Some(callback_wrapper), (&ffi::HW_ENTROPY_DATA) as _, ffi::HW_ENTROPY_LEN as _);
    }
}

pub fn storage_wipe() {
    unsafe {
        ffi::storage_wipe();
    }
}

pub fn storage_unlock(pin: &str) -> bool {
    unsafe {
        let unlocked = ffi::storage_unlock(pin.as_ptr(), pin.len(), ptr::null_mut());
        unlocked == ffi::sectrue
    }
}

pub fn storage_ensure_not_wipe_pin(pin: &str) {
    unsafe {
        ffi::storage_ensure_not_wipe_code(pin.as_ptr(), pin.len());
    }
}

pub fn storage_has_pin() -> bool {
    unsafe {
        ffi::storage_has_pin() == ffi::sectrue
    }
}
pub fn storage_get_remaining() -> u32 {
    unsafe {
        ffi::storage_get_pin_rem()
    }
}

/// Get length of value stored for `appkey`.
/// Returns an error if the value does not exist.
pub fn storage_get_length(appkey: u16) -> StorageResult<usize> {
    let mut len = 0;
    if ffi::sectrue == unsafe { ffi::storage_get(appkey, ptr::null_mut() as _, 0, &mut len) } {
        Ok(len as usize)
    } else {
        Err(StorageError::ReadFailed)
    }
}

/// Get value for `appkey` from storage.
/// The provided `data` buffer must be long enough to hold the value. If the
/// buffer is shorter than the stored value, an error is returned. Use
/// `get_length` to find the size of the value beforehand. On success, the data
/// is copied to `data`, and length of copied data is returned. Data previously
/// in the buffer beyond the copied data is not modified.
/// If `appkey` is not present in storage, or the value is private and storage
/// is locked, returns an error.
pub fn storage_get(appkey: u16, data: &mut [u8]) -> StorageResult<usize> {
    let max_len = data.len().min(u16::MAX as usize) as u16;
    let mut len = 0u16;
    if ffi::sectrue
        == unsafe { ffi::storage_get(appkey, data.as_mut_ptr() as _, max_len, &mut len) }
    {
        Ok(len as usize)
    } else {
        Err(StorageError::ReadFailed)
    }
}

/// Set value for `appkey` in storage.
/// The maximum length of the value is `u16::MAX`. Passing longer buffers will
/// result in an error.
/// If storage is locked and the value is not public-writable, returns an error.
pub fn storage_set(appkey: u16, data: &[u8]) -> StorageResult<()> {
    if data.len() > u16::MAX as usize {
        Err(StorageError::InvalidData)
    } else if ffi::sectrue
        == unsafe { ffi::storage_set(appkey, data.as_ptr() as _, data.len() as u16) }
    {
        Ok(())
    } else {
        Err(StorageError::WriteFailed)
    }
}

/// Set value for monotonic counter `appkey` in storage.
pub fn storage_set_counter(appkey: u16, counter: u32) -> StorageResult<()> {
    if ffi::sectrue == unsafe { ffi::storage_set_counter(appkey, counter) } {
        Ok(())
    } else {
        Err(StorageError::WriteFailed)
    }
}

/// Delete value for `appkey` from storage.
/// Returns true if the value was successfully deleted.
/// If the storage is locked, or the value does not exist in storage, returns
/// false.
pub fn storage_delete(appkey: u16) -> bool {
    ffi::sectrue == unsafe { ffi::storage_delete(appkey) }
}
